# Copyright (c) 2024 The KUTE contributors

import numpy as np
import getpass
import h5py
import openmm

from kute import __version__
from openmm import unit
from typing import Union
import os
_PATHLIKE = Union[str, bytes, os.PathLike]

class PressureTensorReporter(object):
    """
    Custom OpenMM reporter to save the off-diagonal components of the pressure tensor.

    Args:
        file (str): Name of the file to write the pressure tensor to
    """

    def __init__(self, file: _PATHLIKE, reportInterval: int):


        self._out = h5py.File(file, "w")
        self._writer = None
        self._totalSteps = None

        self._reportInterval = reportInterval

    def __del__(self):
        self._out.close()

    def describeNextReport(self, simulation):
        """
        Function to be called by Openmm, gets information about the next report
        this object will generate.
        """
        steps = self._reportInterval - simulation.currentStep%self._reportInterval

        return (steps, True, True, True, False, False)
    

    def report(self, simulation, state):

        if self._writer is None:
            self._writer = PressureTensorWriter(self._out, simulation)

        volume = state.getPeriodicBoxVolume().value_in_unit(unit.meter**3)
        positions = state.getPositions(True).value_in_unit(unit.angstrom)
        velocities = state.getVelocities(True).value_in_unit(unit.angstrom/unit.picosecond)
        forces = state.getForces(True).value_in_unit(unit.kilojoule_per_mole/unit.angstrom)
        time = state.getTime().value_in_unit(unit.picosecond)

        self._writer.writePressureTensor(time, positions, velocities, forces, volume) 


class PressureTensorWriter(object):

    def __init__(self, file: h5py.File, simulation):
        """Class to write electric current in binary h5 format. Helper to PressureTensorReporter

        Args:
            file (h5py.File): File where information will be written
            simulation : OpenMM object describing the simulation
        """
        self._file = file
        self._calculateWeights(simulation)
        self._prepareWriter()

    def _calculateWeights(self, simulation):

        system = simulation.context.getSystem()
        D_TO_KG = 1.66054e-27
        NA = 6.02214076e23
        self._UNITS_FIRST_TERM = D_TO_KG * (1e-10)**2 / ((1e-12)**2) # Units will be J
        self._UNITS_SECOND_TERM = 1_000 / NA # Units will be J

        self._atom_masses = np.array([0.0 for _ in simulation.topology.atoms()])

        ## Set the mass of the residue to which each atom belongs
        for residue in simulation.topology.residues():
            mass = 0
            for atom in residue.atoms():
                mi = system.getParticleMass(atom.index).value_in_unit(unit.dalton)
                mass += mi
                self._atom_masses[atom.index] = mi



    def _prepareWriter(self):

        # Create metadata group

        self._file.create_group('information')
        self._file['information'].attrs['kute_version'] = __version__
        self._file['information'].attrs['openmm_version'] = openmm.Platform.getOpenMMVersion()
        self._file['information'].attrs['author'] = getpass.getuser()
        self._file['information'].create_group('units')
        self._file['information/units'].attrs['time'] = "ps"
        self._file['information/units'].attrs['pressure_tensor'] = "Pa" 

        # create a group for the pressure tensor

        self._file.create_group('timeseries')

        self._file['timeseries'].create_dataset('time', shape=(0, ), maxshape=(None,), dtype=float)
        self._file['timeseries'].create_dataset(f'pressure_tensor', shape=(0, 3), maxshape=(None, 3), dtype=float)

    def writePressureTensor(self, time: float, positions: np.ndarray, velocities: np.ndarray, forces: np.ndarray, volume: float):
        """Write the (center of mass) off-diagonal components of the pressure tensor to the file

        Args:
            time (float): current value of the simulation time
            positions (np.ndarray): Positions of all the atoms in the system
            velocities (np.ndarray): Velocities of all the atoms in the system
            forces (np.ndarray): Forces of all the atoms in the system
        """
        

        pressure_tensor = np.triu(self._UNITS_FIRST_TERM * np.tensordot(self._atom_masses[:, np.newaxis]*velocities, velocities, axes=(0, 0)) + self._UNITS_SECOND_TERM * np.tensordot(positions, forces, axes=(0, 0)), k=1)
        
        off_diagonal = pressure_tensor.flatten()[np.flatnonzero(pressure_tensor)] / volume

        ## Save the results

        LEN = self._file['timeseries/time'].shape[0]
        self._file['timeseries/time'].resize((LEN+1,))
        self._file['timeseries/time'][-1] = time
        self._file[f'timeseries/pressure_tensor'].resize((LEN+1, 3))
        self._file[f'timeseries/pressure_tensor'][-1] = off_diagonal


import argparse
import numpy as np
import MDAnalysis as mda

from kute import GreenKuboIntegral, IntegralEnsemble
from kute.loaders import load_pressure_tensor
from kute.analysis import PressureTensor

def run_one_analysis(top, traj, filename, verbose=True):

    u = mda.Universe(top, traj)
    analysis = PressureTensor(universe=u, filename=filename)
    analysis.run(verbose=verbose)

def calculation_for_one(name, splits, weight=None):
        
    if weight is not None:
        weight = np.loadtxt(weight)
    else:
        weight = 1

    t, Pxy, Pxz, Pyz = load_pressure_tensor(name, splits)
    integral = GreenKuboIntegral(t, Pxy, Pxz, Pyz)
    tavg, avg, u_avg = integral.get_isotropic_running_average()

    avg *= weight
    u_avg *= weight

    return tavg, avg, u_avg

def calculation_for_ensemble(names, splits, weights=None):

    if weights is not None:
        weights = np.loadtxt(weights)

    integrals = []
    for name in names:
        t, Pxy, Pxz, Pyz = load_pressure_tensor(name, splits)
        integrals.append(GreenKuboIntegral(t, Pxy, Pxz, Pyz))
    
    ensemble = IntegralEnsemble(integrals, factors=weights)
    return ensemble.get_isotropic_average()

def save_results(out_file, tavg, avg, uavg):

    to_save = np.vstack([tavg, avg, uavg]).T
    np.savetxt(out_file, to_save, header="Time       Average         Uncertainty")

def main():

    description = 'Calculates the viscosity as a function of averaging cutoff'

    parser = argparse.ArgumentParser(description=description)

## Input arguments

    parser.add_argument("-f", required=False, type=str, dest="h5_files", metavar="pressure.h5", nargs="+", default=None,
                        help = ("List of h5 binary files containing the off-diagonal components of the pressure tensor for each replica."))

    
    parser.add_argument("--traj", required=False, type=str, dest="traj_files", metavar="traj.trr, traj.xtc", nargs="+",
                        help = ("List of trajectory files used to calculate the pressure tensor. Note that they must contain positions, velocities and forces"))


    parser.add_argument("--topo", required=False, type=str, dest="topo_files", metavar="traj.tpr, data.lmp", nargs="+",
                        help = ("List of topology files used to calculate the pressure tensor. They must contain mass information"))
    
    ## Calculation arguments

    parser.add_argument("--splits", required=False, type=int, dest="splits", metavar="slipts", default=1,
                        help = ("Number of fragments of equal size into which to split the pressure tensor of each replica"))

    parser.add_argument("--weights", required=False, type=str, default=None, metavar="weights.dat", dest="weights", 
                        help=("File containing the weighting factor for each replica. Can be used to change units or to include replica-dependent values such as the volume."))

    ## Output arguments

    parser.add_argument("-o", required=False, type=str, dest="out_file", metavar="viscosity.dat", default="viscosity.dat",
                        help = ("Name of the output file"))
    
    args = parser.parse_args()

    ### Check if h5 file was given. If not, carry out the calcultions from the trajectories

    if args.h5_files is None:
        
        print("Didn't specify an h5 file. Falling back to current calculation from MD trajectories")
        for i, (top, traj) in enumerate(zip(args.topo_files, args.traj_files)):
            run_one_analysis(top, traj, f"current_from_traj_replica_{i+1}.h5")

        names = [ f"current_from_traj_replica_{i+1}.h5" for i in range(len(args.topo_files)) ]
    else:
        names = args.h5_files

    ### Carry out the Green-Kubo calculations


    if len(names) == 1:
        tavg, avg, uavg = calculation_for_one(names[0], args.splits, args.weights)
    
    else:
        tavg, avg, uavg = calculation_for_ensemble(names, args.splits, args.weights)

    ## Save the results

    save_results(args.out_file, tavg, avg, uavg)


if __name__ == 'main':
    main()


# Copyright (c) 2024 The KUTE contributors

import numpy as np
import getpass
import h5py
import openmm

from kute import __version__
from openmm import unit
from typing import Union
import os
_PATHLIKE = Union[str, bytes, os.PathLike]

class COMVelocityReporter(object):
    """
    Custom OpenMM reporter to save center of mass velocities

    Args:
        file (str): Name of the file to write the velocities to

    """

    def __init__(self, file: _PATHLIKE, reportInterval: int):
        self._out = h5py.File(file, "w")
        self._reportInterval = reportInterval
        self._writer = None
    
    def __del__(self):
        self._out.close()
    
    def describeNextReport(self, simulation):
        """
        Function to be called by Openmm, gets information about the next report
        this object will generate.
        """

        steps = self._reportInterval - simulation.currentStep%self._reportInterval

        return (steps, False, True, False, False, None)

    def report(self, simulation, state):
        """
        Function to be called by Openmm, generate a report.
        """
        if self._writer is None:

            self._writer = COMVelocityWriter(self._out, simulation)


        velocities = state.getVelocities(True).value_in_unit(unit.angstrom/unit.picosecond)
        time = state.getTime().value_in_unit(unit.picosecond)
        self._writer.writeVelocities(time, velocities)


class COMVelocityWriter(object):
    """Class to write center of mass velocities. Helper to COMVelocityReporter
    Args:
        file (h5py.File): File where information will be written
        simulation : OpenMM object describing the simulation
    """

    def __init__(self, file: h5py.File, simulation):
        self._file = file
        self._calculateWeights(simulation)
        self._prepareWriter(simulation)


    def _calculateWeights(self, simulation):


        system = simulation.context.getSystem()

        self._atom_residuemass = np.array([0.0 for _ in simulation.topology.atoms()])
        self._atom_masses = np.array([0.0 for _ in simulation.topology.atoms()])

        ## Set the mass of the residue to which each atom belongs

        for residue in simulation.topology.residues():

            mass = 0
            for atom in residue.atoms():
                mi = system.getParticleMass(atom.index).value_in_unit(unit.dalton)
                mass += mi
                self._atom_masses[atom.index] = mi

            for atom in residue.atoms():
                self._atom_residuemass[atom.index] = mass

        self._weight_vector = self._atom_masses / self._atom_residuemass
        self._n_atoms = len(self._weight_vector)
        self._n_residues = np.array([1 for _ in simulation.topology.residues()]).sum()
        self._matrix = np.zeros((self._n_residues, self._n_atoms))

        for i, res in enumerate(simulation.topology.residues()):
            for atom in res.atoms():
                j = atom.index
                self._matrix[i, j] = self._weight_vector[j]


    def _prepareWriter(self, simulation):

        # Create metadata group

        self._file.create_group('information')
        self._file['information'].attrs['kute_version'] = __version__
        self._file['information'].attrs['openmm_version'] = openmm.Platform.getOpenMMVersion()
        self._file['information'].attrs['author'] = getpass.getuser()
        self._file['information'].create_group('units')
        self._file['information/units'].attrs['time'] = "ps"
        self._file['information/units'].attrs['com_velocities'] = "A / ps"

        ## Create identificators for which parts corresponds to which residue

        self._file.create_group("residues")
        names = []
        for r in simulation.topology.residues():
            names.append(r.name)
        names = np.array(names)
        for name in np.unique(names):
            where = np.where(names==name)[0]
            self._file['residues'].create_dataset(name, data=where, dtype=int)


        # Create velocity group

        self._file.create_group('timeseries')

        self._file['timeseries'].create_dataset('time', shape=(0, ), maxshape=(None,), dtype=float)
        self._file['timeseries'].create_dataset(f'com_velocities', shape=(0, self._n_residues,3), maxshape=(None, self._n_residues, 3), dtype=float)
        

    def writeVelocities(self, time: float, velocities: np.ndarray):
        """Write the center of mass velocities to the file

        Args:
            time (float): current value of the simulation time
            velocities (np.ndarray): Velocities of all the atoms in the system
        """

        com_velocities = self._matrix @ velocities
        LEN = self._file['timeseries/time'].shape[0]
        self._file['timeseries/time'].resize((LEN+1,))
        self._file['timeseries/time'][-1] = time
        self._file[f'timeseries/com_velocities'].resize((LEN+1, self._n_residues, 3))
        self._file[f'timeseries/com_velocities'][-1] = com_velocities


# Copyright (c) 2024 The KUTE contributors

import numpy as np
import getpass
import openmm
import h5py

from kute import __version__
from openmm import unit
from typing import Union
import os
_PATHLIKE = Union[str, bytes, os.PathLike]

class ElectricCurrentReporter(object):
    """
    Custom OpenMM reporter to save electric current

    Args:
        file (str): Name of the file to write the current to

    """

    def __init__(self, file: _PATHLIKE, reportInterval: int):


        self._out = h5py.File(file, "w")
        self._writer = None
        self._totalSteps = None

        self._reportInterval = reportInterval

    def __del__(self):
        self._out.close()

    def describeNextReport(self, simulation):
        """
        Function to be called by Openmm, gets information about the next report
        this object will generate.
        """
        steps = self._reportInterval - simulation.currentStep%self._reportInterval

        return (steps, False, True, False, False, None)

    def report(self, simulation, state):
        """
        Function to be called by Openmm, generate a report.
        """
        if self._writer is None:
                self._writer = ElectricCurrentWriter(self._out, simulation)


        velocities = state.getVelocities(True).value_in_unit(unit.angstrom/unit.picosecond)
        time = state.getTime().value_in_unit(unit.picosecond)
        self._writer.writeCurrent(time, velocities)


class ElectricCurrentWriter(object):

    def __init__(self, file: h5py.File, simulation):
        """Class to write electric current in binary h5 format. Helper to ElectricCurrentReporter

        Args:
            file (h5py.File): File where information will be written
            simulation : OpenMM object describing the simulation
        """
        self._file = file
        self._calculateWeights(simulation)
        self._prepareWriter()

    def _calculateWeights(self, simulation):

        ## Electric current can be calculated as a sum over all atoms
        ## j = q_i m_i v_i / M_i = (Q_i * m_i / M_i) * v_i = w_i * v_i
        ## Uppercase: residue property. Lowercase: atomic property
        ## we want to compute the weight vector w_i.

        system = simulation.context.getSystem()
        atom_charges = []
        nonbonded = [f for f in system.getForces() if isinstance(f, openmm.NonbondedForce)][0]

        for i in range(system.getNumParticles()):
            charge, _, _ = nonbonded.getParticleParameters(i)
            atom_charges.append(charge)

        atom_charges = np.array(atom_charges)

        self._atom_residuemass = np.array([0.0 for _ in simulation.topology.atoms()])
        self._atom_residuecharge = np.array([0.0 for _ in simulation.topology.atoms()])
        self._atom_masses = np.array([0.0 for _ in simulation.topology.atoms()])

        ## Set the mass of the residue to which each atom belongs

        for residue in simulation.topology.residues():

            mass = 0
            for atom in residue.atoms():
                mi = system.getParticleMass(atom.index).value_in_unit(unit.dalton)
                mass += mi
                self._atom_masses[atom.index] = mi

            for atom in residue.atoms():
                self._atom_residuemass[atom.index] = mass

        ## Set the charge of the residue to which each atom belongs

        for residue in simulation.topology.residues():

            q = 0
            for atom in residue.atoms():
                q += atom_charges[atom.index].value_in_unit(unit.elementary_charge)

            for atom in residue.atoms():
                self._atom_residuecharge[atom.index] = q

        self._weight_vector = self._atom_residuecharge * self._atom_masses / self._atom_residuemass


    def _prepareWriter(self):

        # Create metadata group
    
        self._file.create_group('information')
        self._file['information'].attrs['kute_version'] = __version__
        self._file['information'].attrs['openmm_version'] = openmm.Platform.getOpenMMVersion()
        self._file['information'].attrs['author'] = getpass.getuser()
        self._file['information'].create_group('units')
        self._file['information/units'].attrs['time'] = "ps"
        self._file['information/units'].attrs['electric_current'] = "e * A / ps"

        # Create current group

        self._file.create_group('timeseries')
        self._file['timeseries'].create_dataset('time', shape=(0, ), maxshape=(None,), dtype=float)
        self._file['timeseries'].create_dataset('current', shape=(0, 3), maxshape=(None, 3), dtype=float)


    def writeCurrent(self, time: np.ndarray, velocities:np.ndarray):

        current = np.sum(velocities*self._weight_vector[:, np.newaxis], axis=0)

        LEN = self._file['timeseries/time'].shape[0]
        self._file['timeseries/time'].resize((LEN+1,))
        self._file['timeseries/time'][-1] = time
        self._file['timeseries/current'].resize((LEN+1,3))
        self._file['timeseries/current'][-1] = current



# Copyright (c) 2024 The KUTE contributors

import os 
import pytest
import numpy as np
import kute.__main__
import sys



ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def topo_file():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/topo.tpr")
    return file

@pytest.fixture
def traj_file():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/traj.trr")
    return file

@pytest.fixture
def current_file_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    return file

@pytest.fixture
def velocity_file_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5")
    return file

@pytest.fixture
def pressure_file_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/pressure_tensor_test.h5")
    return file

class TestElectricConductivityRoutine(object):

    def test_run_one_analysis(self, topo_file, traj_file):

        from kute.routines._electric_conductivity import run_one_analysis

        run_one_analysis(topo_file, traj_file, "temporal_result.h5", verbose=False)

        assert os.path.isfile("temporal_result.h5")

        os.remove("temporal_result.h5")

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_calculation_for_one(self, current_file_h5, splits):

        from kute.routines._electric_conductivity import calculation_for_one

        ## Create random weight for the test
        weight = np.array(np.random.random((1,1)))
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_one(current_file_h5, splits, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("splits,repetitions", [(1, 5), (2, 5), (10, 5), (1, 10), (2, 10), (10, 10)])
    def test_calculation_for_ensemble(self, current_file_h5, splits, repetitions):

        from kute.routines._electric_conductivity import calculation_for_ensemble

        weight = np.random.random(repetitions)
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_ensemble([current_file_h5 for _ in range(repetitions)], splits, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("number", [100, 500, 1000])
    def test_save_results(self, number):
    
        from kute.routines._electric_conductivity import save_results

        t, avg, uavg = np.random.random((3, number))
        save_results("temporal.dat", t, avg, uavg)

        assert os.path.isfile("temporal.dat")

        os.remove("temporal.dat")

    @pytest.mark.parametrize("splits,repetitions", [(1, 5), (2, 5), (10, 5), (1, 10), (2, 10), (10, 10)])
    def test_commandline_function(self, current_file_h5, splits, repetitions):

        file_list = " ".join([current_file_h5 for _ in range(repetitions)])
        print(file_list)
        command = f"kute electric_conductivity -f {file_list} --splits {splits} -o temporary_out.dat"
        sys.argv = command.split()

        kute.__main__.main()

        assert os.path.isfile("temporary_out.dat")
        os.remove("temporary_out.dat")


class TestViscosityRoutine(object):

    def test_run_one_analysis(self, topo_file, traj_file):

        from kute.routines._viscosity import run_one_analysis

        run_one_analysis(topo_file, traj_file, "temporal_result.h5", verbose=False)

        assert os.path.isfile("temporal_result.h5")

        os.remove("temporal_result.h5")

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_calculation_for_one(self, pressure_file_h5, splits):

        from kute.routines._viscosity import calculation_for_one

        ## Create random weight for the test
        weight = np.array(np.random.random((1,1)))
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_one(pressure_file_h5, splits, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("splits,repetitions", [(1, 5), (2, 5), (10, 5), (1, 10), (2, 10), (10, 10)])
    def test_calculation_for_ensemble(self, pressure_file_h5, splits, repetitions):
            
            from kute.routines._viscosity import calculation_for_ensemble
    
            weight = np.random.random(repetitions)
            np.savetxt("weight_temp.dat", weight)
    
            t, avg, uavg = calculation_for_ensemble([pressure_file_h5 for _ in range(repetitions)], splits, "weight_temp.dat")
    
            assert len(t) == len(avg)
            assert len(avg) == len(uavg)
    
            os.remove("weight_temp.dat")

    @pytest.mark.parametrize("number", [100, 500, 1000])
    def test_save_results(self, number):
        
            from kute.routines._viscosity import save_results
    
            t, avg, uavg = np.random.random((3, number))
            save_results("temporal.dat", t, avg, uavg)
    
            assert os.path.isfile("temporal.dat")
    
            os.remove("temporal.dat")

    @pytest.mark.parametrize("splits,repetitions", [(1, 5), (2, 5), (10, 5), (1, 10), (2, 10), (10, 10)])
    def test_commandline_function(self, pressure_file_h5, splits, repetitions):
            
            file_list = " ".join([pressure_file_h5 for _ in range(repetitions)])
            command = f"kute viscosity -f {file_list} --splits {splits} -o temporary_out.dat"
            sys.argv = command.split()
    
            kute.__main__.main()
    
            assert os.path.isfile("temporary_out.dat")
            os.remove("temporary_out.dat")


class TestDiffusionCoefficientRoutine(object):

    def test_run_one_analysis(self, topo_file, traj_file):

        from kute.routines._diffusion_coefficient import run_one_analysis

        run_one_analysis(topo_file, traj_file, "temporal_result.h5", verbose=False)

        assert os.path.isfile("temporal_result.h5")

        os.remove("temporal_result.h5")

    @pytest.mark.parametrize("splits,resname", [(1, "ea"), (1, "no3"), (10, "ea"), (10, "no3")])
    def test_calculation_for_one(self, velocity_file_h5, splits, resname):

        from kute.routines._diffusion_coefficient import calculation_for_one

        ## Create random weight for the test
        weight = np.array(np.random.random((1,1)))
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_one(velocity_file_h5, splits, resname, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("splits,repetitions,resname", [(1, 5, "ea"), (1, 5, "no3"), (10, 2, "ea"), (10, 2, "no3")])
    def test_calculation_for_ensemble(self, velocity_file_h5, splits, repetitions, resname):

        from kute.routines._diffusion_coefficient import calculation_for_ensemble

        weight = np.random.random(repetitions)
        np.savetxt("weight_temp.dat", weight)

        t, avg, uavg = calculation_for_ensemble([velocity_file_h5 for _ in range(repetitions)], splits, resname, "weight_temp.dat")

        assert len(t) == len(avg)
        assert len(avg) == len(uavg)

        os.remove("weight_temp.dat")

    @pytest.mark.parametrize("number", [100, 500, 1000])
    def test_save_results(self, number):
    
        from kute.routines._diffusion_coefficient import save_results

        t, avg, uavg = np.random.random((3, number))
        save_results("temporal.dat", t, avg, uavg)

        assert os.path.isfile("temporal.dat")

        os.remove("temporal.dat")

    @pytest.mark.parametrize("splits,repetitions,resname", [(1, 5, "ea"), (1, 5, "no3"), (10, 2, "ea"), (10, 2, "no3")])
    def test_commandline_function(self, velocity_file_h5, splits, repetitions, resname):

        file_list = " ".join([velocity_file_h5 for _ in range(repetitions)])
        command = f"kute diffusion_coefficient -f {file_list} --splits {splits} -o temporary_out.dat --resname {resname}"
        sys.argv = command.split()

        kute.__main__.main()

        assert os.path.isfile("temporary_out.dat")
        os.remove("temporary_out.dat")


import numpy as np
import h5py
import getpass

from kute import __version__
from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis import Universe



class PressureTensor(AnalysisBase):
    """
    Class to calculate the off diagonal components of the pressure tensor from MD trajectories.

    Args:
        universe (MDAnalysis.Universe): Universe containig the simulation
        filename (str, optional): Name of the h5 file to which the pressure will be saved. 
                                  Defaults to "pressure_tensor.h5".
    """

    def __init__(self, universe: Universe, filename: str="pressure_tensor.h5"):
        self.u = universe
        self._trajectory = self.u.trajectory
        self._total_steps = len(universe.trajectory)
        self.filename = filename

    def _prepare(self):

        self._off_diagonal = np.zeros((self._total_steps, 3))

    def _single_frame(self):
        
        positions = self.u.atoms.positions * 1e-10
        velocities = self.u.atoms.velocities * 1e-10 / 1e-12
        forces = self.u.atoms.forces * 1e3 / (1e-10 * 6.02214076e23)
        volume = self.u.dimensions[:3].prod() * 1e-30

        m_v_v_tensor = 1.66054e-27 * np.tensordot(self.u.atoms.masses[:, np.newaxis]*velocities, velocities, axes=(0, 0))

        r_f_tensor = np.tensordot(positions, forces, axes=(0, 0))

        tensor = np.triu(m_v_v_tensor + r_f_tensor, k=1)

        self._off_diagonal[self._frame_index, :] = tensor.flatten()[np.flatnonzero(tensor)] / volume

    def _conclude(self):

        self.write_h5_file()
    
    def write_h5_file(self):
        """
        Write the pressure to an h5 file
        """

        with h5py.File(self.filename, "w") as file:

            ## Metadata group

            file.create_group('information')
            file['information'].attrs['kute_version'] = __version__
            file['information'].attrs['author'] = getpass.getuser()
            file['information'].create_group('units')
            file['information/units'].attrs['time'] = "ps"
            file['information/units'].attrs['pressure_tensor'] = "Pa"

            ## Data group

            file.create_group('timeseries')
            file['timeseries'].create_dataset('time', data = self.times)
            file['timeseries'].create_dataset('pressure_tensor', data = self._off_diagonal)



# Copyright (c) 2024 The KUTE contributors

import numpy as np
from scipy.signal import correlate
from scipy.integrate import cumulative_trapezoid
from typing import Optional, Dict, List, Tuple


class GreenKuboIntegral():
    """
    Class to calculate and manage correlation functions, their integrals and transport properties derived from them.

    Args:
        t (np.ndarray): One dimensional array of shape (N). The timesteps in
            which the current is measured.
        
        J1 (np.ndarray): Two dimensional array of shape (M, N). Each row is an
            independent measurement of N points of the first component of the
            current.
        
        J2 (np.ndarray): Two dimensional array of shape (M, N). Each row is an
            independent measurement of N points of the second component of the
            current.
        
        J3 (np.ndarray): Two dimensional array of shape (M, N). Each row is an
            independent measurement of N points of the third component of the
            current.
    """

    def __init__(self, t: np.ndarray, J1: np.ndarray, J2: np.ndarray, J3: np.ndarray):
        self.time = t
        self.J1 = J1
        self.J2 = J2
        self.J3 = J3
        # self.caf: Dict[str, Optional[np.ndarray]] = {
        #     "11": None, "22": None,"33": None, "12": None,"13": None,
        #     "23": None}
        self.caf: Dict[str, np.ndarray] = {}
        # self.cumulative_integral: Dict[str, Optional[np.ndarray]] = {
        #     "11": None, "22": None, "33": None, "12": None, "13": None,
        #     "23": None}
        self.cumulative_integral: Dict[str, np.ndarray] = {}
        # self.running_average: Dict[str, Optional[np.ndarray]] = {
        #     "11": None, "22": None, "33": None, "12": None, "13": None,
        #     "23": None}
        self.running_average: Dict[str, np.ndarray] = {}
        self.dt: float = t[1] - t[0]
        self._calculated_caf = False
        self._calculated_integral = False
        self._calculated_running_average = False

        
    def set_time_to_zero(self):
        """
        Shifts the time origin so that self.time[0] = 0
        """
        self.time -= self.time[0] * np.ones(len(self.time))


    @staticmethod
    def get_correlation_function(j1: np.ndarray, j2: np.ndarray) -> np.ndarray:
        """
        Method for quickly calculating a correlation function and its uncertainty.

        Args:
            j1 (np.ndarray): Two dimensional array of shape (M, N). It represents M measurings of a signal consisting of
            N points.
            j2 (np.ndarray): Two dimensional array of shape (M, N). It represents M measurings of a signal consisting of
            N points.

        Returns:
            np.ndarray: Two dimensional array of shape (N-1, 2). The first
            column contains the correlation <j1(t0) * j2(t0+t)>, averaged over
            all time origins t0 and all samples M The second column contains the
            uncertainty of the correlation function.
        
        Raises:
            ValueError: If the two entries don't have the same shape
        """

        if j1.shape != j2.shape:
            raise ValueError("Correlation cannot be calculated between entries with different sizes")

        try:
            M, N = j1.shape
        except:
            j1 = j1[np.newaxis, :]
            j2 = j2[np.newaxis, :]
            M, N = j1.shape


        caf = np.zeros(N-1)
        caf2 =np.zeros(N-1)
        contributions = np.array([N-k for k in range(N-1)])

        for A in range(M):

            caf += correlate(j1[A, :], j2[A, :])[N-1:-1]
            caf2 += correlate(j1[A, :]**2, j2[A, :]**2)[N-1:-1]

        caf /= M * contributions
        caf2 /= M * contributions

        unc_caf = np.sqrt(caf2 - caf**2) / np.sqrt(M*contributions - 1)

        return np.array([caf, unc_caf]).T
    
    def _calculate_correlation(self):    
        """
        Calculates all independent autocorrelation functions, C_11, C_22, C_33,
        C_12, C_13 and C_23, as well as their uncertainties. They are stored in
        self.caf under different keys.
        """ 
        
        if not self._calculated_caf:

            self.caf["11"] = self.get_correlation_function(self.J1, self.J1)
            self.caf["22"] = self.get_correlation_function(self.J2, self.J2)
            self.caf["33"] = self.get_correlation_function(self.J3, self.J3)
            self.caf["12"] = self.get_correlation_function(self.J1, self.J2)
            self.caf["13"] = self.get_correlation_function(self.J1, self.J3)
            self.caf["23"] = self.get_correlation_function(self.J2, self.J3)

            self._calculated_caf = True
        

    def _calculate_integral(self):
        """
        Calculates the cumulative integral of the various autocorrelation
        functions and their uncertainties. They get stored in
        self.cumulative_integral under different keys.
        """
        
        if not self._calculated_caf:
            raise RuntimeError("The correlation functions have not been calculated yet")

        if not self._calculated_integral:

            for key in list(self.caf.keys()):
                caf = self.caf[key][:, 0]
                unc_caf = self.caf[key][:, 1]
                cumul = cumulative_trapezoid(caf, self.time[:-1], initial = 0)
                
                unc_trapzs = np.zeros(len(caf))
                unc_trapzs[1:] = (self.dt/2)*np.sqrt(np.array([ unc_caf[i]**2 + unc_caf[i-1]**2 for i in range(1, len(cumul)) ]))
                unc_cumul = np.sqrt(np.cumsum(unc_trapzs**2))
                self.cumulative_integral[key] = np.array([cumul, unc_cumul]).T
            
            self._calculated_integral = True

           

    
    def _calculate_running_average(self):
        """
            Calculates the running average for all time steps, given by the
            weighted average of the cumulative integral from a given timestep to
            the last one, and for all tensor components. This has N-1 values,
            with N the length of the cumulative integral. The values and their
            uncertainties are stored in self.running_average under different
            keys.
        """

        if not self._calculated_running_average:
            if not self._calculated_caf or not self._calculated_integral:
                raise RuntimeError("The correlation functions have not been calculated yet")

            for key in list(self.caf.keys()):

                ## The values of the cumulative integral and the corresponding weights are stored in arrays
                ## The arrays are flipped so that they structure is [ cumul(N), cumul(N-1), ..., cumul(0) ]

                flipped_cumul = np.flip(self.cumulative_integral[key][1:, 0])
                flipped_weights = np.flip(self.cumulative_integral[key][1:, 1]) ** (-2)
                contributions = np.flip(np.arange(len(flipped_cumul)-1) + 2)

                ## From the flipped arrays, the weighted average for each choice of beggining of averaging can be calculated as follows
                ## The last point is not calculated, because it corresponds to an average of a single point

                running_average = np.flip(np.cumsum(flipped_cumul * flipped_weights) / np.cumsum(flipped_weights))[:-1]
                running_average2 = np.flip(np.cumsum(flipped_cumul**2 * flipped_weights) / np.cumsum(flipped_weights))[:-1]

                uncertainty = np.sqrt((running_average2 - running_average**2)/contributions)

                shape = (len(running_average), 2)
                self.running_average[key] = np.zeros(shape)
                self.running_average[key][:, 0] = running_average
                self.running_average[key][:, 1] = uncertainty
            
            self._calculated_running_average = True
        


    def analyze(self):
        """
            Performs the analysis for the system. This consists of calculating
            correlations, running integrals and running averages.
        """

        self._calculate_correlation()
        self._calculate_integral()
        self._calculate_running_average()

    
    def get_running_average(self, keys:list)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the running average integral and its uncertainty, averaging
        over the given components of the transport tensor
        Args:
            keys (list): A selection of keys ("11", "22", "33", "12", "13", "23") to average over

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """

        self.analyze()
        NKEYS = len(keys)
        L = len(self.running_average[keys[0]][:, 0])
        values = np.zeros((NKEYS, L))
        weights = np.zeros((NKEYS, L))

        for j, key in enumerate(keys):

            values[j, :] = self.running_average[key][:, 0]
            weights[j, :] = self.running_average[key][:, 1] ** (-2)

        average = np.sum(values*weights, axis=0) / np.sum(weights, axis=0)
        variance2 = np.sum(weights * (values - average[np.newaxis, :])**2, axis=0) / np.sum(weights, axis=0)
        uncertainty = np.sqrt((1./(NKEYS-1)) * variance2)

        return self.time[:L], average, uncertainty


    def get_isotropic_running_average(self)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the running average integral and its uncertainty, averaging
        over the isotropic components of the transport tensor

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """
        
        return self.get_running_average(["11", "22", "33"])
        

    def get_anisotropic_running_average(self)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the running average integral and its uncertainty, averaging
        over the anisotropic components of the transport tensor

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """

        return self.get_running_average(["12", "13", "23"])


    def get_average_correlation_function(self, keys):
        """
        Calculates the running correlation function and its uncertainty, averaging
        over the given components of the transport tensor
        Args:
            keys (list): A selection of keys ("11", "22", "33", "12", "13", "23") to average over

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Average correlation function
            uncertainty (np.ndarray): Uncertainty of the correlation function
        """
        self.analyze()
        NKEYS = len(keys)
        L = len(self.caf[keys[0]][:, 0])
        values = np.zeros((NKEYS, L))
        weights = np.zeros((NKEYS, L))

        for j, key in enumerate(keys):

            values[j, :] = self.caf[key][:, 0]
            weights[j, :] = self.caf[key][:, 1] ** (-2)

        average = np.sum(values*weights, axis=0) / np.sum(weights, axis=0)
        variance2 = np.sum(weights * (values - average[np.newaxis, :])**2, axis=0) / np.sum(weights, axis=0)
        uncertainty = np.sqrt((1./(NKEYS-1)) * variance2)

        return self.time[:L], average, uncertainty
    

    def get_isotropic_correlation_function(self):
        """
        Calculates the average correlation function and its uncertainty, averaging
        over the isotropic components

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic correlation function
            uncertainty (np.ndarray): Uncertainty of the correlation function
        """
        return self.get_average_correlation_function(["11", "22", "33"])
    
    def get_anisotropic_correlation_function(self):
        """
        Calculates the average correlation function and its uncertainty, averaging
        over the anisotropic components

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic correlation function
            uncertainty (np.ndarray): Uncertainty of the correlation function
        """
        return self.get_average_correlation_function(["12", "13", "23"])


class IntegralEnsemble():
    """
    Class to manage a collection of replicas when calculating transport
    properties

    Args:
        green_kubo_integrals (list): A list of instances of the
            GreenKuboIntegral class, with the values of the current for each
            replica
        factors (np.ndarray,optional): Multiplying factors for the average
            integral of each replica. Can be used for changing units, or to
            implement a factor that changes among
            replicas (1/V for example). The default is no weights are applied.
    Raises
    ------
    ValueError 
        If the GreenKuboIntegral objects have non matching times
    ValueError
        If the length of factors does not match the number of replicas
    """

    def __init__(self, green_kubo_integrals: List[GreenKuboIntegral], 
                 factors: Optional[np.ndarray]=None):
        self.gk_integrals = green_kubo_integrals
        self.N_REPLICAS = len(green_kubo_integrals)
        self.time = green_kubo_integrals[0].time
        self.LENGTH = len(self.time) - 3

        for integral in green_kubo_integrals:
            if np.all(self.time != integral.time):
                raise ValueError("Integrals with non matching times were given")
            
        if factors is None:
            self.FACTORS = np.ones(self.N_REPLICAS)
        else:
            if len(factors) != self.N_REPLICAS:
                raise ValueError("Length of the prefactors does not match the number of replicas")
            else:
                self.FACTORS = factors


    def get_average_over_components(self, keys: list) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the running average of one or more components of the transport tensor across all replicas

        Args:
            key (list): A list of strings (11, 22, 33, 12, 13 or 23). The components to be 
                        averaged over

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """

        replica_values = np.zeros((self.N_REPLICAS, self.LENGTH))
        replica_weights = np.zeros((self.N_REPLICAS, self.LENGTH))

        for i, (integral, FACTOR) in enumerate(zip(self.gk_integrals, self.FACTORS)):
            
            integral.analyze()
            _, val, uval = integral.get_running_average(keys)
            replica_values[i, :] = FACTOR * val
            replica_weights[i, :] = FACTOR * uval

        average = np.sum(replica_values*replica_weights, axis=0) / np.sum(replica_weights, axis=0)
        variance2 = np.sum(replica_weights * (replica_values - average[np.newaxis, :])**2, axis=0) / np.sum(replica_weights, axis=0)
        uncertainty = np.sqrt(variance2 / (self.N_REPLICAS - 1))

        return self.time[:self.LENGTH], average, uncertainty
    

    def get_isotropic_average(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the isotropic average of the transport tensor across all replicas

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """

        return self.get_average_over_components(["11", "22", "33"])
    

    def get_anisotropic_average(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the anisotropic average of the transport tensor across all replicas

        Returns:
            time (np.ndarray): Times corresponding to the values
            average (np.ndarray): Isotropic running average
            uncertainty (np.ndarray): Uncertainty of the running average
        """

        return self.get_average_over_components(["12", "13", "23"])







# Copyright (c) 2024 The KUTE contributors

import numpy as np
import h5py
import getpass

from kute import __version__
from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis import Universe


class ElectricCurrent(AnalysisBase):
    """
    Class to calculate electric currents from MD trajectories.

    Args:
        universe (MDAnalysis.Universe): Universe containig the simulation
        filename (str, optional): Name of the h5 file to which the current will be saved. 
                                  Defaults to "current.h5".
    """
    
    def __init__(self, universe: Universe, filename: str="current.h5"):
        self.u = universe
        self._trajectory = self.u.trajectory
        self._total_steps = len(universe.trajectory)
        self.filename = filename

    def _prepare(self):
        atom_masses = self.u.atoms.masses
        residue_masses = np.array([ a.residue.mass for a in self.u.atoms ])
        residue_charges = np.array([ a.residue.charge for a in self.u.atoms ])

        self._weights = residue_charges * atom_masses / residue_masses
        
        self._current = np.zeros((self._total_steps, 3))

    def _single_frame(self):

        self._current[self._frame_index, :] = np.sum(self.u.atoms.velocities * self._weights[:, np.newaxis], axis=0)

    def _conclude(self):

        self.write_h5_file()


    def write_h5_file(self):
        """
        Write the current to an h5 file
        """

        with h5py.File(self.filename, "w") as file:

            ## Metadata group

            file.create_group('information')
            file['information'].attrs['kute_version'] = __version__
            file['information'].attrs['author'] = getpass.getuser()
            file['information'].create_group('units')
            file['information/units'].attrs['time'] = "ps"
            file['information/units'].attrs['electric_current'] = "e * A / ps"

            ## Data group

            file.create_group('timeseries')
            file['timeseries'].create_dataset('time', data = self.times)
            file['timeseries'].create_dataset('current', data = self._current)



# Copyright (c) 2024 The KUTE contributors

import os 
import pytest
import openmm
import getpass 
import h5py
import openmm.app as app
import numpy as np

from openmm import unit
from kute import __version__
from kute.reporters import ElectricCurrentReporter, COMVelocityReporter, PressureTensorReporter
from kute.loaders import load_electric_current, load_com_velocity, load_pressure_tensor


ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def simulation():

    FIELD = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/field_openmm.xml")
    CONFIG = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/config_openmm.pdb")
    
    TEMP = 300 * unit.kelvin
    TIMESTEP = 1.0 * unit.femtosecond

    forcefield = app.ForceField(FIELD)
    pdb = app.PDBFile(CONFIG)

    system = forcefield.createSystem(pdb.topology,
                                     nonbondedMethod=app.PME,
                                     nonbondedCutoff=12.0*unit.angstrom,
                                     constraints=app.HBonds,
                                     ewaldErrorTolerance=1.0e-5)

    integrator = openmm.NoseHooverIntegrator(TEMP, 10 / unit.picoseconds, TIMESTEP)
    sim = app.Simulation(pdb.topology, system, integrator)
    sim.context.setPositions(pdb.positions)
    sim.context.setVelocitiesToTemperature(TEMP)

    return sim


class TestElectricCurrentReporter(object):

    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        current_reporter = ElectricCurrentReporter(os.path.join(ACTUAL_PATH, "report_current_temp.h5"), reportInterval=1)
        simulation.reporters.append(current_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_current_temp.h5")
        assert os.path.isfile(FILEPATH)

        ## Check if the file can be loaded with KUTE

        t, Jx, Jy, Jz = load_electric_current(FILEPATH)
        
        assert Jx.shape == (1, len(t))
        assert Jy.shape == (1, len(t))
        assert Jz.shape == (1, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH, "r") as f:

            # Check general information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["electric_current"] == "e * A / ps"

            # Check for correct shapes

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/current"]).shape == (TOTALSTEPS, 3)

        os.remove(FILEPATH)

class TestCOMVelocityReporter(object):

    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        vel_reporter = COMVelocityReporter(os.path.join(ACTUAL_PATH, "report_com_vel_temp.h5"), reportInterval=1)
        simulation.reporters.append(vel_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_com_vel_temp.h5")
        assert os.path.isfile(FILEPATH)

        # Check if it can be loaded with KUTE

        names = np.unique([r.name for r in simulation.topology.residues()])
        for name in names:
            t, Vx, Vy, Vz = load_com_velocity(FILEPATH, name)
            number = np.sum([ 1 if r.name==name else 0 for r in simulation.topology.residues() ])
            assert Vx.shape == (number, len(t))
            assert Vy.shape == (number, len(t))
            assert Vz.shape == (number, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH) as f:

            # General information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["com_velocities"] == "A / ps"

            # Check residue counts

            for name in names:

                number = np.sum([ 1 if r.name==name else 0 for r in simulation.topology.residues() ])
                assert len(np.array(f[f"residues/{name}"])) == number

            # Check the shape of the results
            
            target_shape = (TOTALSTEPS, simulation.topology.getNumResidues(), 3)

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/com_velocities"]).shape == target_shape

        
        os.remove(FILEPATH)

class TestPressureTensorReporter(object):


    def test_reporter(self, simulation):

        simulation.reporters.clear()
        
        current_reporter = PressureTensorReporter(os.path.join(ACTUAL_PATH, "report_pressure_tensor_temp.h5"), reportInterval=1)
        simulation.reporters.append(current_reporter)

        TOTALSTEPS = 100
        simulation.step(TOTALSTEPS)

        # Check if report exists

        FILEPATH = os.path.join(ACTUAL_PATH, "report_pressure_tensor_temp.h5")
        assert os.path.isfile(FILEPATH)

        ## Check if the file can be loaded with KUTE

        t, Pxy, Pxz, Pyz = load_pressure_tensor(FILEPATH)
        
        assert Pxy.shape == (1, len(t))
        assert Pxz.shape == (1, len(t))
        assert Pyz.shape == (1, len(t))

        # Checks for the h5 file

        with h5py.File(FILEPATH, "r") as f:

            # Check general information

            assert f["information"].attrs["kute_version"] == __version__
            assert f["information"].attrs["author"] == getpass.getuser()
            assert f["information/units"].attrs["time"] == "ps"
            assert f["information/units"].attrs["pressure_tensor"] == "Pa"

            # Check for correct shapes

            assert len(np.array(f["timeseries/time"])) == TOTALSTEPS
            assert np.array(f["timeseries/pressure_tensor"]).shape == (TOTALSTEPS, 3)

        os.remove(FILEPATH)


# Copyright (c) 2024 The KUTE contributors

import os
import pytest
import h5py
import numpy as np
from kute.loaders import load_com_velocity, load_electric_current, load_pressure_tensor


ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def current_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    return h5py.File(file)

@pytest.fixture
def com_velocity_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5")
    return h5py.File(file)

@pytest.fixture
def pressure_tensor_h5():
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/pressure_tensor_test.h5")
    return h5py.File(file)

@pytest.fixture
def current(splits):

    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/current_test.h5")
    return load_electric_current(file, splits)

@pytest.fixture
def com_velocity(splits, resname):

    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5")
    return load_com_velocity(file, resname, splits)

@pytest.fixture
def pressure_tensor(splits):
    file = os.path.join(ACTUAL_PATH, "files_for_tests/inputs/pressure_tensor_test.h5")
    return load_pressure_tensor(file, splits)


class TestElectricCurrent(object):

    @pytest.mark.parametrize("splits", [(1)])
    def test_load(self, current_h5, current):

        time, Jx, Jy, Jz = current
        time_h5 = np.array(current_h5["timeseries/time"])
        Jx_h5, Jy_h5, Jz_h5 = np.array(current_h5["timeseries/current"]).T

        assert np.allclose(time, time_h5)
        assert np.allclose(Jx, Jx_h5)
        assert np.allclose(Jy, Jy_h5)
        assert np.allclose(Jz, Jz_h5)
    

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_shape(self, current_h5, current, splits):

        time, Jx, Jy, Jz = current
        len_h5 = len(np.array(current_h5["timeseries/time"]))
        target_len = len_h5 / splits

        ## Check for correct splitting of the time
        assert (len(time) == target_len)

        ## Check that all currents have the correct shape

        correct_shape = (splits, target_len)

        assert Jx.shape == correct_shape
        assert Jy.shape == correct_shape
        assert Jz.shape == correct_shape

class TestPressureTensor(object):

    @pytest.mark.parametrize("splits", [(1)])
    def test_load(self, pressure_tensor_h5, pressure_tensor):

        time, Px, Py, Pz = pressure_tensor
        time_h5 = np.array(pressure_tensor_h5["timeseries/time"])
        Px_h5, Py_h5, Pz_h5 = np.array(pressure_tensor_h5["timeseries/pressure_tensor"]).T

        assert np.allclose(time, time_h5)
        assert np.allclose(Px, Px_h5)
        assert np.allclose(Py, Py_h5)
        assert np.allclose(Pz, Pz_h5)

    @pytest.mark.parametrize("splits", [1, 2, 5, 10])
    def test_shape(self, pressure_tensor_h5, pressure_tensor, splits):

        time, Px, Py, Pz = pressure_tensor
        len_h5 = len(np.array(pressure_tensor_h5["timeseries/time"]))
        target_len = len_h5 / splits

        ## Check for correct splitting of the time
        assert (len(time) == target_len)

        ## Check that all currents have the correct shape

        correct_shape = (splits, target_len)

        assert Px.shape == correct_shape
        assert Py.shape == correct_shape
        assert Pz.shape == correct_shape


class TestCOMVelocity(object):

    @pytest.mark.parametrize("splits,resname", [(1, "ea"), (1, "no3"), (2, "ea"), (2, "no3"), (5, "ea"), (5, "no3"), (10, "ea"), (10, "no3")])
    def test_load(self, com_velocity_h5, com_velocity, splits, resname):


        time, Vx, Vy, Vz = com_velocity
        time_h5 = np.array(com_velocity_h5["timeseries/time"])
        
        if resname == "ea":
            Vx_h5, Vy_h5, Vz_h5 = np.split(np.array(com_velocity_h5["timeseries/com_velocities"]), 2, axis=1)[0].T

        elif resname == "no3":
            Vx_h5, Vy_h5, Vz_h5 = np.split(np.array(com_velocity_h5["timeseries/com_velocities"]), 2, axis=1)[1].T

        Vx_h5 = np.vstack(np.split(Vx_h5, splits, axis=1))
        Vy_h5 = np.vstack(np.split(Vy_h5, splits, axis=1))
        Vz_h5 = np.vstack(np.split(Vz_h5, splits, axis=1))

        assert np.allclose(time, time_h5[:len(time_h5)//splits])
        assert np.allclose(Vx, Vx_h5)
        assert np.allclose(Vy, Vy_h5)
        assert np.allclose(Vz, Vz_h5)

    @pytest.mark.parametrize("splits", [1, 5, 10])
    def test_missing_residue(self, splits):
        with pytest.raises(ValueError):
            _, _, _, _ = load_com_velocity(os.path.join(ACTUAL_PATH, "files_for_tests/inputs/com_velocity_test.h5"), "missing", splits)


    @pytest.mark.parametrize("splits,resname", [(1, "ea"), (1, "no3"), (2, "ea"), (2, "no3"), (5, "ea"), (5, "no3"), (10, "ea"), (10, "no3")])
    def test_shape(self, com_velocity_h5, com_velocity, splits, resname):

        _, Vx, Vy, Vz = com_velocity
        time_h5 = np.array(com_velocity_h5["timeseries/time"])

        n_residues = len(np.array(com_velocity_h5[f"residues/{resname}"]))

        time_h5 = time_h5[:len(time_h5)//splits]
        target_shape = (n_residues * splits, len(time_h5))

        assert Vx.shape == target_shape
        assert Vy.shape == target_shape
        assert Vz.shape == target_shape




# Copyright (c) 2024 The KUTE contributors

__version__ = "1.0.5"

from ._kute import GreenKuboIntegral, IntegralEnsemble
from . import analysis
from . import loaders

__all__ = ["GreenKuboIntegral", "IntegralEnsemble", "analysis", "loaders"]


# Copyright (c) 2024 The KUTE contributors

from ._comvelocityreporter import COMVelocityReporter
from ._electriccurrentreporter import ElectricCurrentReporter
from ._pressuretensorreporter import PressureTensorReporter

__all__ = ["COMVelocityReporter", "ElectricCurrentReporter", "PressureTensorReporter"]


import sys
import kute.routines._diffusion_coefficient as diffusion_coefficient
import kute.routines._electric_conductivity as electric_conductivity
import kute.routines._viscosity as viscosity

routines = {
    "diffusion_coefficient": diffusion_coefficient,
    "electric_conductivity": electric_conductivity,
    "viscosity": viscosity
}
def main():
    STRING_AVAILABLE = f"Available routines: {list(routines.keys())}"
    if len(sys.argv) == 1:
        print(STRING_AVAILABLE)
        return
    if sys.argv[1] in routines:
        sys.argv[0] = "kute " + sys.argv[1]
        name = sys.argv.pop(1)
        routines[name].main()
    else:
        print(STRING_AVAILABLE)
        sys.exit(1)
        
if __name__ == "__main__":
    main()